--
-- DrivingLine
-- Specialization for driving lines of sowing machines
--
--	@author:		gotchTOM
--	@date: 			19-May-2013
--	@version: 		v1.19beta
--	@history:		v1.0 	- initial implementation (17-Jun-2012)
--					v1.19beta
--						 	- replace limitToField
--							- manualDrivingLine	
--							- movable HUD
--							- compatible with ZZZ_HiredWorkerConsumesFuelSeeds.zip
--							- self.backUpMoveDown
--							- attacherVehicle bugfix
--							- mousecursor detach bugfix	
--							- camera movement bugfix
--							- mousecursor AI problem bugfix	
--							- compatible with Kuhn Moduliner (self.drivemark)
--							- compatible with Farmet_Excelent_Premium_6 (self.seederIsDown)
--							- LS 2013 compatible
--							- getRootAttacherVehicle
--							- new HUD & some bugfixes concerning mousecursor/camera, improved switching of the tramlines in different sowing machines 
--							- expanded number of possible spray widths
--							- setTextAlignment


DrivingLine = {};

function DrivingLine.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(SowingMachine, specializations);
end;

function DrivingLine:load(xmlFile)

	self.setDrivingLine = SpecializationUtil.callSpecializationsFunction("setDrivingLine");
	self.setSPworkwidth = SpecializationUtil.callSpecializationsFunction("setSPworkwidth");
	local numdrivingLines = Utils.getNoNil(getXMLInt(xmlFile, "vehicle.drivingLines#count"), 0);
	self.drivingLines = {}
    for i=1, numdrivingLines do
        self.drivingLines[i] = {};
        local areanamei = string.format("vehicle.drivingLines.drivingLine" .. "%d", i); 
        self.drivingLines[i].start = Utils.indexToObject(self.components, getXMLString(xmlFile, areanamei .. "#startIndex"));
        self.drivingLines[i].width = Utils.indexToObject(self.components, getXMLString(xmlFile, areanamei .. "#widthIndex"));
        self.drivingLines[i].height = Utils.indexToObject(self.components, getXMLString(xmlFile, areanamei .. "#heightIndex"));
    end;
	if numdrivingLines > 0 then
		self.drivingLinePresent = true;
	end;
	self.drivingLineActiv = false;
	self.dlHUDisActive = false;
	self.manualDrivingLine = true;
	self.backUpMoveDown = false;
	
	self.dlAllowUpdate = true;
	
	self.currentDrive = 1;
	self.smWorkwith = 3;
	self.nSMdrives = 2;
	self.dlWarning = 0;
	self.dlCultivatorDelay = 0;
	
	
	-- HUD
	--local overlayfile = Utils.getFilename("textures/dlHUD.png", self.baseDirectory);
	local overlayfile = Utils.getFilename("textures/amazone_dl.dds", self.baseDirectory);
	self.dlOverlay = createImageOverlay(overlayfile);
	setOverlayColor(self.dlOverlay, 1, 1, 1, 1);
	
	self.dlHudBaseX = 0.845;
	self.dlHudBaseY = 0.59;
	self.dlHudMove = false;
	self.mouseEnabled = false;
end;

function DrivingLine:delete()
end;

function DrivingLine:readStream(streamId, connection)

	self.drivingLineActiv = streamReadBool(streamId);
end;

function DrivingLine:writeStream(streamId, connection)

	streamWriteBool(streamId, self.drivingLineActiv);
end;

function DrivingLine:mouseEvent(posX, posY, isDown, isUp, button)
	
	-- HUD move
	if self.dlHudMove then
		self.dlHudBaseX = posX;
		self.dlHudBaseY = posY;
	end;
	local dlHudchangedJet = false;
	
	if isDown and button == 1 then
		if self.dlHudMove then
			self.dlHudMove = false;
			dlHudchangedJet = true;
		end;
		if self.dlHUDisActive then
			-- move HUD
			if not dlHudchangedJet then
				if (self.dlHudBaseX-0.10) < posX and posX < (self.dlHudBaseX+0.2) and (self.dlHudBaseY-0.01) < posY and posY < (self.dlHudBaseY+0.02) then
					self.dlHudMove = true;
				end;
			end;
			-- switch manual-automatic
			if (self.dlHudBaseX+0.14) > posX and (self.dlHudBaseX+0.10) < posX and (self.dlHudBaseY-0.01) > posY and (self.dlHudBaseY-0.04) < posY then
				self.manualDrivingLine = not self.manualDrivingLine;
			end;
			if not self.manualDrivingLine then
				-- set current drive
				if (self.dlHudBaseX+0.12) > posX and (self.dlHudBaseX+0.10) < posX and (self.dlHudBaseY-0.04) > posY and (self.dlHudBaseY-0.07) < posY then
					self.currentDrive = self.currentDrive + 1;
				end;
				if (self.dlHudBaseX+0.13) > posX and (self.dlHudBaseX+0.12) < posX and (self.dlHudBaseY-0.04) > posY and (self.dlHudBaseY-0.07) < posY then
					self.currentDrive = self.currentDrive - 1;
				end;
				-- set workwith of sowing machine
				if (self.dlHudBaseX+0.12) > posX and (self.dlHudBaseX+0.10) < posX and (self.dlHudBaseY-0.10) > posY and (self.dlHudBaseY-0.13) < posY then
					if self.smWorkwith < 15 then
						self.smWorkwith = self.smWorkwith + 1;
					else
						self.smWorkwith = 3;
						self.dlWarning = self.time + 250;
					end;
					self.dlAllowUpdate = true;					
				end;
				if (self.dlHudBaseX+0.13) > posX and (self.dlHudBaseX+0.12) < posX and (self.dlHudBaseY-0.10) > posY and (self.dlHudBaseY-0.13) < posY then
					if self.smWorkwith > 3 then
						self.smWorkwith = self.smWorkwith - 1;
					else
						self.smWorkwith = 15;
						self.dlWarning = self.time + 250;
					end;
					self.dlAllowUpdate = true;
				end;
				-- set workwith of sprayer 
				if (self.dlHudBaseX+0.12) > posX and (self.dlHudBaseX+0.10) < posX and (self.dlHudBaseY-0.13) > posY and (self.dlHudBaseY-0.16) < posY then
					self:setSPworkwidth(true);
				end;
				if (self.dlHudBaseX+0.13) > posX and (self.dlHudBaseX+0.12) < posX and (self.dlHudBaseY-0.13) > posY and (self.dlHudBaseY-0.16) < posY then
					self:setSPworkwidth(false);
				end;
			end;	
		end;
	end;	
end;

function DrivingLine:setSPworkwidth(raise)

	if not raise then
		if self.nSMdrives > 2 and self.spWorkwith < 61 then
			self.nSMdrives = self.nSMdrives - 1;
		else
			self.nSMdrives = 2;
			self.dlWarning = self.time + 250;
		end;
	else
		if self.spWorkwith < 57 then
			self.nSMdrives = self.nSMdrives + 1;
		else
			self.nSMdrives = 2;
			self.dlWarning = self.time + 250;
		end;
	end;
	
	self.dlAllowUpdate = true;
end;

function DrivingLine:keyEvent(unicode, sym, modifier, isDown)
end;

function DrivingLine:onDetach(attacherVehicle)
    
	if self.mouseEnabled then
		InputBinding.setShowMouseCursor(false);
	end;
end;

function DrivingLine:onLeave()

	if self.mouseEnabled then
		InputBinding.setShowMouseCursor(false);
	end
end

function DrivingLine:update(dt)

	if self:getIsActiveForInput() then
		if self.drivingLinePresent then				
			-- switch driving line
			if InputBinding.hasEvent(InputBinding.DRIVINGLINE) then
				if self.drivingLineActiv then
					self:setDrivingLine(false);
				else
					self:setDrivingLine(true);
				end;
			end;
			-- switch HUD
			if InputBinding.hasEvent(InputBinding.DRIVINGLINE_HUD) then
				self.dlHUDisActive = not self.dlHUDisActive;
				self.mouseEnabled = self.dlHUDisActive;
				if not self.mouseEnabled then
					InputBinding.setShowMouseCursor(false);
				end;
			end;
			-- switch mouse cursor
			if self.dlHUDisActive then
				if InputBinding.hasEvent(InputBinding.DRIVINGLINE_MOUSECURSOR) then
					if self.mouseEnabled then
						self.mouseEnabled = false;
						InputBinding.setShowMouseCursor(false);
					else
						self.mouseEnabled = true;
					end;
				end;
			end;
		end;	
	end;
end;

function DrivingLine:updateTick(dt)
	
	if self:getIsActive() then
		if self.drivingLinePresent then
			if self.isServer then
				if self.drivingLineActiv and self.dlCultivatorDelay <= self.time then
					local allowCultivator = false;
					if self.needsActivation then
						if (self.sowingMachineHasGroundContact or self.seederIsDown) and self.isTurnedOn then
							allowCultivator = true;
						end;
					else
						if self.sowingMachineHasGroundContact then
							allowCultivator = true;						
						end;
					end;
					if allowCultivator then
						local drivingLinesSend = {};
						for _,area in pairs(self.drivingLines) do
							local x,y,z = getWorldTranslation(area.start);
							local x1,y1,z1 = getWorldTranslation(area.width);
							local x2,y2,z2 = getWorldTranslation(area.height);
							table.insert(drivingLinesSend, {x,z,x1,z1,x2,z2});
							if table.getn(drivingLinesSend) > 0 then
								CultivatorAreaEvent.runLocally(drivingLinesSend, true);
								g_server:broadcastEvent(CultivatorAreaEvent:new(drivingLinesSend, true));
							end;
						end;
					end;				
				end;
			end;	
			if not self.manualDrivingLine then	
			
				if self.currentDrive > self.nSMdrives then
					self.currentDrive = 1;
				elseif self.currentDrive < 1 then
					self.currentDrive = self.nSMdrives;
				end;
				
				if self.dlAllowUpdate then
					self.spWorkwith = self.smWorkwith * self.nSMdrives;
					if (self.nSMdrives%2 == 0) then -- gerade Zahl
						self.num_DrivingLine = (self.nSMdrives / 2) + 1;
					elseif (self.nSMdrives%2 ~= 0) then -- ungerade Zahl
						self.num_DrivingLine = (self.nSMdrives + 1) / 2;	
					end;
					self.dlAllowUpdate = false;
				end;
				
				if self.currentDrive == self.num_DrivingLine and not self.drivingLineActiv then
					self:setDrivingLine(true);
				elseif self.currentDrive ~= self.num_DrivingLine and self.drivingLineActiv then
					self:setDrivingLine(false);
				end;
				
				local rootAttacherVehicle = self:getRootAttacherVehicle();
				if rootAttacherVehicle ~= nil then		
					local implement = rootAttacherVehicle.selectedImplement;
					if implement ~= nil then				
						local jointDesc = rootAttacherVehicle.attacherJoints[implement.jointDescIndex];
						local moveDown = jointDesc.moveDown;
						if self.backUpMoveDown ~= moveDown then
							if not jointDesc.moveDown then
								self.currentDrive = self.currentDrive + 1;
								self.dlCultivatorDelay = self.time + 2000;
							end;
							self.backUpMoveDown = moveDown;								
						end;	
					end;
				end;
			end;
		end;	
    end;
end;

function DrivingLine:setIsTurnedOn()
    
	local rootAttacherVehicle = self:getRootAttacherVehicle();
	if rootAttacherVehicle ~= nil then	
		if not self.allowsLowering	and not self.manualDrivingLine and not self.isTurnedOn and (rootAttacherVehicle.isControlled or rootAttacherVehicle.isHired) then
			self.currentDrive = self.currentDrive + 1;
		end;
	end;	
end;

function DrivingLine:draw()

	if self.drivingLinePresent then
		if self.manualDrivingLine then
			if self.drivingLineActiv then	
				g_currentMission:addHelpButtonText(g_i18n:getText("DRIVINGLINE_OFF"), InputBinding.DRIVINGLINE);
			else
				g_currentMission:addHelpButtonText(g_i18n:getText("DRIVINGLINE_ON"), InputBinding.DRIVINGLINE);
			end;
		end;
		-- show mouse cursor
		if self.mouseEnabled then
			InputBinding.setShowMouseCursor(true);
		end;
		
		if self.dlHUDisActive then
			g_currentMission:addHelpButtonText(g_i18n:getText("DRIVINGLINE_HUD_OFF"), InputBinding.DRIVINGLINE_HUD);
			-- if self.mouseEnabled then
				-- g_currentMission:addHelpButtonText(g_i18n:getText("DRIVINGLINE_MOUSECURSOR"), InputBinding.DRIVINGLINE_MOUSECURSOR);
			-- else
				-- g_currentMission:addHelpButtonText(g_i18n:getText("DRIVINGLINE_MOUSECURSOR"), InputBinding.DRIVINGLINE_MOUSECURSOR);
			-- end;
			if self.dlWarning > self.time then
				setTextColor(1,0.2,0.2,1);
			else
				setTextColor(0.15,0.15,0.15,1);
			end;
			setTextBold(false);
			setTextAlignment(RenderText.ALIGN_LEFT); 
			--renderText((self.dlHudBaseX-0.10), (self.dlHudBaseY-0.01), 0.025, g_i18n:getText("DRIVINGLINE_MOVEHUD"));
			renderText((self.dlHudBaseX+0.105), (self.dlHudBaseY-0.04), 0.025, "< >");
			if not self.manualDrivingLine then
				renderText((self.dlHudBaseX-0.105), (self.dlHudBaseY-0.04), 0.025, g_i18n:getText("DRIVINGLINE_AUTOMATIC"));
				renderText((self.dlHudBaseX-0.105), (self.dlHudBaseY-0.07), 0.025, g_i18n:getText("DRIVINGLINE_CURRENTDRIVE").." "..tostring(self.currentDrive).." ("..tostring(self.nSMdrives)..")");
				
				renderText((self.dlHudBaseX+0.105),(self.dlHudBaseY-0.07), 0.025, "+ -");
				renderText((self.dlHudBaseX-0.105), (self.dlHudBaseY-0.10), 0.02, g_i18n:getText("DRIVINGLINE_NUMDRIVINGLINE").." "..tostring(self.num_DrivingLine));
				
				
				renderText((self.dlHudBaseX-0.105), (self.dlHudBaseY-0.13), 0.025, g_i18n:getText("DRIVINGLINE_WORKWIDTHSOWING").." "..tostring(self.smWorkwith));
				renderText((self.dlHudBaseX+0.105),(self.dlHudBaseY-0.13), 0.025, "+ -");
				
				renderText((self.dlHudBaseX-0.105), (self.dlHudBaseY-0.16), 0.025, g_i18n:getText("DRIVINGLINE_WORKWIDTHSPRAYER").." "..tostring(self.spWorkwith));
				renderText((self.dlHudBaseX+0.105),(self.dlHudBaseY-0.16), 0.025, "+ -");
				
			else
				renderText((self.dlHudBaseX-0.105), (self.dlHudBaseY-0.04), 0.025, g_i18n:getText("DRIVINGLINE_MANUAL"));
				renderText((self.dlHudBaseX-0.105), (self.dlHudBaseY-0.07), 0.025, g_i18n:getText("DRIVINGLINE")..":");
				renderText((self.dlHudBaseX-0.105), (self.dlHudBaseY-0.10), 0.025, InputBinding.getKeyNamesOfDigitalAction(InputBinding.DRIVINGLINE));

			end;
			renderOverlay(self.dlOverlay, (self.dlHudBaseX-0.125), (self.dlHudBaseY-0.355), 0.284, 0.4);				
			
		else
			g_currentMission:addHelpButtonText(g_i18n:getText("DRIVINGLINE_HUD_ON"), InputBinding.DRIVINGLINE_HUD);
		end;	
		-- setTextColor(1,1,1,1);
	end;
end;

function DrivingLine:setDrivingLine(drivingLineActiv, noEventSend)

	if noEventSend == nil or noEventSend == false then	
		SetDrivingLineEvent.sendEvent(self, drivingLineActiv, noEventSend); 
	end;	
	self.drivingLineActiv = drivingLineActiv;
	
	
	-- Kuhn Moduliner
	if self.drivemark ~= nil then
		self.drivemark = self.drivingLineActiv;
	end;
			

end;



SetDrivingLineEvent = {};
SetDrivingLineEvent_mt = Class(SetDrivingLineEvent, Event);

InitEventClass(SetDrivingLineEvent, "SetDrivingLineEvent");

function SetDrivingLineEvent:emptyNew()
  
    local self = Event:new(SetDrivingLineEvent_mt);
    self.className="SetDrivingLineEvent";
    return self;
end;

function SetDrivingLineEvent:new(vehicle, drivingLineActiv)

    local self = SetDrivingLineEvent:emptyNew()
    self.vehicle = vehicle;
	self.drivingLineActiv = drivingLineActiv;
    return self;
end;

function SetDrivingLineEvent:readStream(streamId, connection) 

    local id = streamReadInt32(streamId); 
	self.drivingLineActiv = streamReadBool(streamId);  
    self.vehicle = networkGetObject(id); 
    self:run(connection);
	-- print("readStream! self.drivingLineActiv: "..tostring(self.drivingLineActiv).." self.vehicle: "..tostring(self.vehicle).." streamId: "..tostring(streamId).." connection: "..tostring(connection)) --!!!

end;

function SetDrivingLineEvent:writeStream(streamId, connection) 
  
    streamWriteInt32(streamId, networkGetObjectId(self.vehicle));	
	streamWriteBool(streamId, self.drivingLineActiv);
	-- print("writeStream! self.drivingLineActiv: "..tostring(self.drivingLineActiv).." self.vehicle: "..tostring(self.vehicle).." streamId: "..tostring(streamId).." connection: "..tostring(connection)) --!!!

end;

function SetDrivingLineEvent:run(connection) 
  
	if not connection:getIsServer() then	
		for k, v in pairs(g_server.clientConnections) do
			if v ~= connection and not v:getIsLocal() then
				v:sendEvent(SetDrivingLineEvent:new(self.vehicle, self.drivingLineActiv));
			end;
		end;
	end;	
	self.vehicle:setDrivingLine(self.drivingLineActiv, true); 
	-- print("SetDrivingLineEvent:run(connection)") --!!!
end;

function SetDrivingLineEvent.sendEvent(vehicle, drivingLineActiv, noEventSend)  
		
	if g_server ~= nil then  
		g_server:broadcastEvent(SetDrivingLineEvent:new(vehicle, drivingLineActiv), nil, nil, vehicle);
		-- print("sendEvent: g_server:broadcast Event! drivingLineActiv: "..tostring(drivingLineActiv).." vehicle: "..tostring(vehicle).." noEventSend: "..tostring(noEventSend)) --!!!
		
	else  
		g_client:getServerConnection():sendEvent(SetDrivingLineEvent:new(vehicle, drivingLineActiv));
		-- print("sendEvent: g_client:send Event! drivingLineActiv: "..tostring(drivingLineActiv).." vehicle: "..tostring(vehicle).." noEventSend: "..tostring(noEventSend)) --!!!
		
	end;
end;